/*
 * Unisphere -- UNIformly distributing points on SPHERE
 * 
 * Copyright (C) 2010 by Ali Katanforoush <a_katanforosh@sbu.ac.ir>
 *
 * This file is a part of Unisphere.
 *
 * Unisphere is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Unisphere is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
// unisphere.cpp

#include "unisphere.h"

static int icoli[12][5]={{1,2,3,4,5},
                         {2,0,5,9,8},
                         {0,1,8,7,3},
                         {0,2,7,11,4},
                         {0,3,11,10,5},
                         {10,9,1,0,4},
                         {7,8,9,10,11},
                         {2,8,6,11,3},
                         {9,6,7,2,1},
                         {1,5,10,6,8},
                         {6,9,5,4,11},
                         {6,10,4,3,7}};

static int icofi[12][5]={{0,1,2,3,4},
                         {0,4,5,6,7},
                         {0,7,8,9,1},
                         {1,9,10,11,2},
                         {2,11,12,13,3},
                         {14,5,4,3,13},
                         {15,16,17,18,19},
                         {8,15,19,10,9},
                         {16,15,8,7,6},
                         {5,14,17,16,6},
                         {17,14,13,12,18},
                         {18,12,11,10,19}};

static double icog[12][3]={{0,1.61803398874989,1},
                           {1.61803398874989,1,0},
                           {0,1.61803398874989,-1},
                           {-1.61803398874989,1,0},
                           {-1,0,1.61803398874989},
                           {1,0,1.61803398874989},
                           {0,-1.61803398874989,-1},
                           {-1,0,-1.61803398874989},
                           {1,0,-1.61803398874989},
                           {1.61803398874989,-1,0},
                           {0,-1.61803398874989,1},
                           {-1.61803398874989,-1,0}};

void cs3rand( int n, double *x, double *y, double *z) {
    double w,r,teta;
    int i;
    for(i=0; i<n; ++i) {
        w=RAN; teta=2*PI*RAN;
        r=2*sqrt(w*(1-w));
        x[i]=r*cos(teta);
        y[i]=r*sin(teta);
        z[i]=1-2*w;
    }
}

void cs4rand( int n, double *x, double *y, double *z, double *t) {
    double r1,r2;
    double teta1, teta2;
    int i;
    for(i=0; i<n; ++i) {
        double w=RAN;
        r1=sqrt(w); teta1=2*PI*RAN;
        r2=sqrt(1-w); teta2=2*PI*RAN;
        x[i]=r1*cos(teta1);
        y[i]=r1*sin(teta1);
        z[i]=r2*cos(teta2);
        t[i]=r2*sin(teta2);
    }
}

void ps3rand( int n, double *teta, double *fi) {
    for(int i=0; i<n; ++i) {
        teta[i]=2*PI*RAN;
        fi[i]=asin(2*RAN-1);
    }
}

void cb2rand( int n, double *x, double *y) {
    double r,teta;
    for(int i=0; i<n; ++i) {
        teta=2*PI*RAN;
        r=sqrt(RAN);
        x[i]=r*cos(teta);
        y[i]=r*sin(teta);
    }
}

void pb2rand( int n, double *r, double *teta) {
    for(int i=0; i<n; ++i) {
        teta[i]=2*PI*RAN;
        r[i]=sqrt(RAN);
    }
}

void tf2xyz(int n, double *teta, double *fi, double *x, double *y, double *z) {
    for(int i=0; i<n; ++i) {
        double c=cos(fi[i]);
        z[i]=sin(fi[i]);
        y[i]=c*sin(teta[i]);
        x[i]=c*cos(teta[i]);
    }
}

int eqdistre(int n, double e, double *teta, double *fi) {
    double nq3=n*sqrt(3);

    double f,df;
    double t,dt;
    int N,k;

    N=0;
    f=-PI/2;
    df=(PI-e)/n;

    for(int i=0; i<=n; ++i) {
        k=floor(nq3*cos(f)+0.5);
        if(k<1) k=1;
        dt=2*PI/k;
        t= (i%2==0) ? 0:dt/2;
        while(k>0) {
            teta[N]=t;
            fi[N]=f;
            --k;
            ++N;
            t+=dt;
        }
        f+=df;
    }

    return N;
}

int NRE(int n, double e) {
    int N=0;
    double nq3=n*sqrt(3);
    double f=-PI/2;
    double df=(PI-e)/n;

    for(int i=0; i<=n; ++i) {
        int k=floor(nq3*cos(f)+0.5);
        if(k<1) k=1;
        N+=k;
        f+=df;
    }

    return N;
}

int interpolateRE(int N, int &n, double &e) {
    int N1,N2;
    int n1,n2;

    e=0;
    n1=floor(sqrt(N/C2));
    n2=n1+1;

    N1=N2=0;
    while(1) {
        if(!N1) N1=NRE(n1,0);
        if(N==N1) {
            n=n1;
            return N;
        }
        if(!N2) N2=NRE(n2,0);
        if(N==N2) {
            n=n2;
            return N;
        }

        if(N<N1) {
            n1=2*n1-n2;
            n2=(n1+n2)/2;
            N2=N1;
            N1=0;
        }
        else if(N2<N) {
            n2=2*n2-n1;
            n1=(n1+n2)/2;
            N1=N2;
            N2=0;
        }
        else {
            if(n2-n1>1) {
                n1=(n1+n2)/2;
                N1=0;
            }
            else
                break;
        }
    }

    double s=PI/n2;
    double t=s+PI/n1;
    double e1,e2,e3;
    e1=0;
    e2=t;

    double qN=sqrt(N);
    double qN1=sqrt(N1);
    double qN2=sqrt(N2);

    int N3;
    for(int r=12; r>0; --r) {
        if(r%3==0)
            e3=((N2-N)*e1+(N-N1)*e2)/(N2-N1);
        else if(r%3==1)
            e3=((qN2-qN)*e1+(qN-qN1)*e2)/(qN2-qN1);
        else
            e3=(0.7*e1+1.3*e2)/2;

        if(e3<s) {
            n=n1;
            e=e3;
        }
        else {
            n=n2;
            e=e3-t;
        }
        N3=NRE(n,e);
        if(N3==N)
            return N;
        else if(N3<N) {
            N1=N3;
            e1=e3;
            qN1=sqrt(N3);
        }
        else {
            N2=N3;
            e2=e3;
            qN2=sqrt(N3);
        }
    }
    return N-N1<N2-N ? N1:N2;
}

int eqdistre(int N, double *teta, double *fi) {
    int n;
    double e;
    interpolateRE(N, n, e);
    return eqdistre(n, e, teta, fi);
}

void resphere(int n, double *x, double *y, double *z) {
    for(int i=n; i>0; --i) {
        double d=sqrt(*x**x+*y**y+*z**z);
        *x/=d;
        *y/=d;
        *z/=d;
        ++x; ++y; ++z;
    }
}

int icolattice(int m, int n, double *x, double *y, double *z) {
    int fvisit[20]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    int evisit[12][12]={{0,0,0,0,0,0,0,0,0,0,0,0},
                        {0,0,0,0,0,0,0,0,0,0,0,0},
                        {0,0,0,0,0,0,0,0,0,0,0,0},
                        {0,0,0,0,0,0,0,0,0,0,0,0},
                        {0,0,0,0,0,0,0,0,0,0,0,0},
                        {0,0,0,0,0,0,0,0,0,0,0,0},
                        {0,0,0,0,0,0,0,0,0,0,0,0},
                        {0,0,0,0,0,0,0,0,0,0,0,0},
                        {0,0,0,0,0,0,0,0,0,0,0,0},
                        {0,0,0,0,0,0,0,0,0,0,0,0},
                        {0,0,0,0,0,0,0,0,0,0,0,0},
                        {0,0,0,0,0,0,0,0,0,0,0,0}};

    int mn=m*m+n*n+m*n;

    int J=0;
    for(int i=0; i<12; ++i) {
        double x0=icog[i][0];
        double y0=icog[i][1];
        double z0=icog[i][2];

        x[J]=x0;
        y[J]=y0;
        z[J]=z0;
        ++J;

        for(int j=0; j<5; ++j) {
            if(fvisit[icofi[i][j]]==0) {
                int j1=icoli[i][j];
                int j2=icoli[i][(j+1)%5];
                int ij1=evisit[i][j1];
                int ij2=evisit[i][j2];
                int j1j2=evisit[j1][j2];
                // ij1=0; ij2=0; j1j2=0;

                double dx1=icog[j1][0]-x0;
                double dy1=icog[j1][1]-y0;
                double dz1=icog[j1][2]-z0;

                double dx2=icog[j2][0]-x0;
                double dy2=icog[j2][1]-y0;
                double dz2=icog[j2][2]-z0;

                double vx1=((m+n)*dx1-n*dx2)/mn;
                double vy1=((m+n)*dy1-n*dy2)/mn;
                double vz1=((m+n)*dz1-n*dz2)/mn;

                double vx2=(n*dx1+m*dx2)/mn;
                double vy2=(n*dy1+m*dy2)/mn;
                double vz2=(n*dz1+m*dz2)/mn;

                double rl1=(double)mn/(double)(m+n);
                double l1=floor(rl1);

                if( j1j2 && rl1==l1 )
                    --l1;

                int l;
                for(l=1; l<=(int)l1; ++l) {
                    x[J]=x0+l*vx2;
                    y[J]=y0+l*vy2;
                    z[J]=z0+l*vz2;
                    ++J;
                }

                int k;
                for(k=1; k<=m-1; ++k) {
                    double rl0=(double)k*n/(double)m;
                    double rl1=(double)(mn-k*m)/(double)(m+n);
                    double l0=ceil(rl0);
                    double l1=floor(rl1);
                    if( ij1 && rl0==l0 )
                        ++l0;
                    if( j1j2 && rl1==l1 )
                        --l1;

                    for(int l=(int)l0; l<=(int)l1; ++l) {
                        x[J]=x0+k*vx1+l*vx2;
                        y[J]=y0+k*vy1+l*vy2;
                        z[J]=z0+k*vz1+l*vz2;
                        ++J;
                    }
                }

                for(k=-n+1; k<=-1; ++k) {
                    double rl0=(double)-k*(m+n)/(double)n;
                    double rl1=(double)(mn-k*m)/(double)(m+n);
                    double l0=ceil(rl0);
                    double l1=floor(rl1);
                    if( ij2 && rl0==l0 )
                        ++l0;
                    if( j1j2 && rl1==l1 )
                        --l1;
                    for(l=(int)l0; l<=(int)l1; ++l) {
                        x[J]=x0+k*vx1+l*vx2;
                        y[J]=y0+k*vy1+l*vy2;
                        z[J]=z0+k*vz1+l*vz2;
                        ++J;
                    }
                }

                evisit[i][j1]=1; evisit[j1][i]=1;
                evisit[i][j2]=1; evisit[j2][i]=1;
                evisit[j2][j1]=1; evisit[j1][j2]=1;
                fvisit[icofi[i][j]]=1;
            }
        }
    }
    return J;
}

int cs3icolattice(int m, int n, double *x, double *y, double *z) {
    int N=icolattice(m,n,x,y,z);
    resphere(N, x,y,z);
    return N;
}

